//---------------------------------------------------------------------------
//
//   SeparateVertex
//
//          Copyright(C) 2013, Tark
//
//    _
//
//---------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include <math.h>
#include <unordered_map>
#include <unordered_set>
#include <algorithm>
#include "MQPlugin.h"
#include "CMQObject.h"
#include "SeparateVertex.h"

//---------------------------------------------------------------------------
//  MQGetPlugInID
//---------------------------------------------------------------------------
void SeparateVertex::GetPlugInID(DWORD *Product, DWORD *ID)
{
	*Product = 0xE3F7A830;
	*ID      = 0xBD9A186E;
}

//---------------------------------------------------------------------------
//  MQGetPlugInName
//---------------------------------------------------------------------------
const char *SeparateVertex::GetPlugInName(void)
{
	return "SeparateVertex ver.1.0 Copyright(C) 2013, Tark.";
}

//---------------------------------------------------------------------------
//  MQEnumString
//---------------------------------------------------------------------------
const char *SeparateVertex::EnumString(int index)
{
	switch(index){
	case 0: return "SeparateVertex"; break;
	case 1: return "SeparateLine"; break;
	}
	return nullptr;
}

MQBasePlugin *GetPluginClass()
{
	static SeparateVertex plugin;
	return &plugin;
}

BOOL SeparateVertex::Execute(int index, MQDocument doc_)
{
	doc = cmq::CMQDocument(doc_);
	obj = doc.getCurrentObject();
	switch(index){
	case 0:
		return separateVertex();
	case 1:
		return separateLine();
	}
	return FALSE;
}

SeparateVertex::SeparateVertex() : doc(nullptr)
{
}

SeparateVertex::~SeparateVertex()
{
}

BOOL SeparateVertex::separateVertex()
{
	cmq::CMQContainer<cmq::CMQVertex, cmq::CMQObject> vcontainer(&obj);
	for(auto vit = vcontainer.begin(); vit != vcontainer.end(); ++vit) {
		if(!vit->isSelect()) {
			continue;
		}
		auto farray = vit->getVertexRelatedFaces();
		if(farray.size() < 2) {
			continue;
		}
		int fno = 0;
		for(cmq::findex_t findex : farray) {
			if(fno++ == 0) {
				// 1͊̂܂
				continue;
			}
			// ǉ_ɍ킹ʂ͍蒼
			cmq::vindex_t newvindex = obj.addVertex(vit->p());
			obj.copyVertexAttribute(newvindex, obj, vit.index());
			auto varray = obj.getFacePointArray(findex);
			auto findit = find(varray.begin(), varray.end(), vit.index());
			if(findit == varray.end()) {
				continue; // whats wrong happend
			}
			*findit = newvindex; // ǉ_ɒu
			// ʍč쐬
			cmq::findex_t newfindex = obj.modifyFace(findex, varray);
		}
	}
	return TRUE;
}


void SeparateVertex::make_maps(
	std::unordered_set<cmq::vindex_t> &vertexset,
	std::unordered_set<cmq::lindex_t> &lineset,
	std::unordered_multimap<cmq::vindex_t, cmq::lindex_t> &vtolmap,
	std::unordered_multimap<cmq::lindex_t, std::pair<cmq::findex_t, int>> &ltofmap)
{
	cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		auto larray = fit->getFacePointArray();
		for(int lindex = 0; lindex < larray.size(); lindex++) {
			auto line = obj.getLineVertex(fit.index(), lindex);
			cmq::lindex_t lineid = cmq::makeLineId(line);
			if(doc.isSelectLine(obj, fit.index(), lindex)) {
				lineset.insert(lineid);
				vertexset.insert(line.first);
				vertexset.insert(line.second);
				vtolmap.insert(std::make_pair(line.first, lineid));
				vtolmap.insert(std::make_pair(line.second, lineid));
			}
			ltofmap.insert(std::make_pair(lineid, std::make_pair(fit.index(), lindex)));
		}
	}
}

BOOL SeparateVertex::dist_groups(std::unordered_map<cmq::vindex_t, std::unordered_multimap<int, cmq::findex_t>> &groups,
	std::unordered_set<cmq::vindex_t> &vertexset,
	std::unordered_set<cmq::lindex_t> &lineset,
	std::unordered_multimap<cmq::lindex_t, std::pair<cmq::findex_t, int>> &ltofmap)
{
	for(cmq::vindex_t vindex : vertexset) {
		auto farray = obj.getVertexRelatedFaces(vindex);
		std::unordered_map<cmq::findex_t, int> ftogmap;
		int groupno = 0;
		for(auto findex :farray) {
			// אږʂTĊɃO[vĂ邩ׂ
			auto larray = obj.getLineArray(findex, vindex); // ڒ_܂ޗאڕ
			if(larray.size() != 2) {
				// something wrong
				return FALSE;
			}
			// IӂŗאڂĂʂO[vĂ炻groupnofoundgroup̏W߂
			std::unordered_set<int> foundgroup;
			for(cmq::lindex_t lid : larray) {
				if(lineset.find(lid) != lineset.end()) {
					// IӂŗאڂĂʂ͌Ȃ
					continue;
				}
				// Yӂ܂ޖʂ̂ftogmapɓo^Ă̂Afoundgroupɂgroupno
				auto itpair = ltofmap.equal_range(lid);
				for(auto it = itpair.first; it != itpair.second; ++it) {
					if(it->second.first != findex) {
						auto findit = ftogmap.find(it->second.first);
						if(findit != ftogmap.end()) {
							foundgroup.insert(findit->second);
						}
					}
				}
			}
			if(foundgroup.empty()) {
				// O[vꂽאږʂ͂Ȃ
				ftogmap.insert(std::make_pair(findex, groupno++));
			}
			else {
				// ɃO[vꂽאږʂAVO[vԍőS㏑
				for(auto it = ftogmap.begin(); it != ftogmap.end(); ++it) {
					if(foundgroup.find(it->second) != foundgroup.end()) {
						it->second = groupno;
					}
				}
				ftogmap.insert(std::make_pair(findex, groupno++));
			}
		}
		std::unordered_multimap<int, cmq::findex_t> gtofmap;
		std::unordered_set<int> groupkey;
		for(auto p : ftogmap) {
			gtofmap.insert(std::make_pair(p.second, p.first));
			groupkey.insert(p.second);
		}
		if(groupkey.size() < 2) {
			// אږʂIӂŕO[vɕĂȂ̂ŁA_Ȃ
			continue;
		}
		groups.insert(make_pair(vindex, gtofmap));
	}
	return TRUE;
}

BOOL SeparateVertex::make_faces(std::unordered_map<cmq::vindex_t, std::unordered_multimap<int, cmq::findex_t>> &groups)
{
	std::unordered_map<cmq::findex_t, std::vector<cmq::vindex_t>> newfaces;
	for(auto vgpair : groups) {
		int vindex = vgpair.first;
		auto gtofmap = vgpair.second;
		int count = 0;
		std::unordered_set<int> groupkey;
		for(auto p : gtofmap) {
			groupkey.insert(p.first);
		}
		cmq::CMQVertex v(obj.getVertex(vindex), &obj, vindex);
		for(int groupno : groupkey) {
			if(count++ == 0) {
				// ڂ̃O[v͊̂܂
				continue;
			}
			// ǉ_ɍ킹ʂ͍蒼
			cmq::vindex_t newvindex = obj.addVertex(v);
			obj.copyVertexAttribute(newvindex, obj, vindex);
			auto findrange = gtofmap.equal_range(groupno);
			for(auto it = findrange.first; it != findrange.second; ++it) {
				auto newfaceit = newfaces.find(it->second);
				if(newfaceit == newfaces.end()) {
					auto varray0 = obj.getFacePointArray(it->second);
					newfaceit = newfaces.insert(make_pair(it->second, varray0)).first;
				}
				auto varray = newfaceit->second;
				auto findit = find(varray.begin(), varray.end(), vindex);
				if(findit == varray.end()) {
					// something wrong
					return FALSE;
				}
				*findit = newvindex; // ǉ_ɒu
				newfaces[it->second] =  varray;
			}
		}
	}
	for(auto fpair : newfaces) {
		cmq::findex_t findex = fpair.first;
		std::vector<cmq::vindex_t> varray = fpair.second;
		// ʍč쐬
		cmq::findex_t newfindex = obj.modifyFace(findex, varray);
	}
	return TRUE;
}

BOOL SeparateVertex::separateLine()
{
	// IӂƑIӒ[_unordered_set
	// _Iӂւunordered_multimap
	// ӂאږʂւunordered_multimap
	std::unordered_set<cmq::vindex_t> vertexset;
	std::unordered_set<cmq::lindex_t> lineset;
	std::unordered_multimap<cmq::vindex_t, cmq::lindex_t> vtolmap;
	std::unordered_multimap<cmq::lindex_t, std::pair<cmq::findex_t, int>> ltofmap;
	make_maps(vertexset, lineset, vtolmap, ltofmap);

	// IӒ[_ɂāAאږʂIӂŋ؂ꂽO[vɕ
	std::unordered_map<cmq::vindex_t, std::unordered_multimap<int, cmq::findex_t>> groups;

	if(!dist_groups(groups, vertexset, lineset, ltofmap)) {
		return FALSE;
	}

	// groupsɉĒ_ǉĖʂ蓖Ē
	return make_faces(groups);
}
